'* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *'
'* SOKOBAN GAME                                     Version 1A *'
'*                                                             *'
'* Written by Zsolt Nagy Perge in Dec. 1999. <zsnp500@cox.net> *'
'* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *'

DEFINT A-Z
'* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *'
'* SUB & FUNCTION procedure definitions                        *'
'* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *'
DECLARE SUB MSAVE (File$)
DECLARE SUB MLOAD (File$)
DECLARE SUB OpenMaze (File$, CHANGE)
DECLARE SUB SaveMaze (File$, CHANGE)
DECLARE SUB DrawCell (x, Y, D)
DECLARE SUB DrawMaze ()
DECLARE SUB ClearMaze ()
DECLARE SUB Initialize ()
DECLARE SUB PRINTC (S$)
DECLARE SUB PutBox (Col1, Row1, Col2, Row2, LineSet$)
DECLARE SUB PutWindow (Col1, Row1, Col2, Row2, LineSet$)
DECLARE SUB ReadScreen (Buffer())
DECLARE SUB WriteScreen (Buffer())
DECLARE SUB SokoPlay ()
DECLARE SUB SokoEdit ()
DECLARE FUNCTION SokoMenu ()
DECLARE FUNCTION GetText (Text$, MaxLength, Filter$, ESCAPE)
DECLARE FUNCTION FindFile (Name$(), Size$())
DECLARE FUNCTION EXCHECK (F$)
DECLARE FUNCTION GETKEY$ ()
DECLARE FUNCTION ASK (Title$, Question$, BC)
'* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *'
'* Global definitions                                          *'
'* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *'
CONST uSPC = 0          ' SPaCe
CONST uWLL = 1          ' WaLL
CONST uBOX = 2          ' BOX
CONST uTRG = 3          ' box TaRGet
CONST uBOT = 4          ' Box On Target
COMMON SHARED EC        ' Run-Time Error Code
COMMON SHARED MBC       ' Maze Box Counter
COMMON SHARED MTC       ' Maze Box Target Counter
COMMON SHARED MWC       ' Maze Wall Counter
COMMON SHARED FP&       ' File pointer
COMMON SHARED CX        ' Character position (horizontal)
COMMON SHARED CY        ' Character position (vertical)
DIM SHARED MAZE$(1 TO 20)
'* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *'
'* Main Program                                                *'
'* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *'
Main:
'    ON ERROR GOTO ErrorHandler

    Initialize

    DO
        SELECT CASE SokoMenu
            CASE 1: SokoPlay: KILL "SOKOBAN.TMP"
            CASE 2: SokoEdit
            CASE 3: EXIT DO
            CASE ELSE:
        END SELECT
    LOOP

    COLOR 7, 0
    LOCATE , , 1
    CLS
    PRINT "Thank you for playing SOKOBAN - The Maze Game! (Version 1A)"
    PRINT "Written by Zsolt Nagy Perge in Dec. 1999."
    PRINT
    SYSTEM
'* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *'
ErrorHandler:
    SCREEN 0
    WIDTH 80, 25
    COLOR 14, 4
    CLS
    BEEP
    PRINT "ERR ="; ERR
    PRINT
    COLOR 15
    PRINT "Please report this error to: zsnp500@cox.net"
    PRINT "Tell me what did you do before this error occurred"
    PRINT "and also send me the value of ERR (above)."
    PRINT "Thank you..."
    PRINT
    SYSTEM
'* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *'

' Ask a "yes or no" question and return the answer...
FUNCTION ASK (Title$, Question$, BC)
    DIM S(4000)
    ReadScreen S()

    Title$ = MID$(Title$, 1, 70)
    Question$ = MID$(Question$, 1, 70)
    TL = LEN(Title$)
    QL = LEN(Question$)

    IF TL > QL THEN WWidth = TL ELSE WWidth = QL

    COLOR 15, BC
    PutWindow 30 - WWidth / 2, 11, 50 + WWidth / 2, 15, "ɻȼͺ*"
    LOCATE 11: PRINTC " " + Title$ + " "
    COLOR 14
    LOCATE 13: PRINTC Question$ + " (Y/N)?"
    Key$ = GETKEY$
    IF UCASE$(Key$) = "Y" OR Key$ = CHR$(13) THEN ASK = 1 ELSE ASK = 0
    WriteScreen S()
END FUNCTION

' Clear the maze...
SUB ClearMaze
    FOR I = 1 TO 20
        MAZE$(I) = STRING$(40, CHR$(0))
    NEXT I

    MBC = 0
    MTC = 0
    MWC = 0
END SUB

' Draws one cell on the screen.
SUB DrawCell (x, Y, D)

    LOCATE Y, x * 2 - 1
    SELECT CASE D
        CASE uSPC:              ' SPACE
            COLOR 7, 0
            PRINT "  ";
        CASE uWLL:              ' WALL
            COLOR 4, 7
            PRINT "";
        CASE uBOX:              ' BOX
            COLOR 15, 1
            PRINT "[]";
        CASE uTRG:              ' BOX TARGET
            COLOR 10, 0
            PRINT "><";
        CASE uBOT:              ' BOX ON TARGET
            COLOR 15, 2
            PRINT "OK";
        CASE ELSE:
    END SELECT
END SUB

' Draws the maze from MAZE$() on the screen.
SUB DrawMaze
    FOR V = 1 TO 20
        FOR H = 1 TO 40
            DrawCell H, V, ASC(MID$(MAZE$(V), H, 1))
        NEXT H
    NEXT V
END SUB

' Checks whether a file exists or not.
' Returns 1 if file F$ exists, otherwise zero.
FUNCTION EXCHECK (F$)
    H = FREEFILE
    OPEN F$ FOR BINARY AS H
    IF LOF(H) > 0 THEN
        CLOSE H
        EXCHECK = 1
    ELSE
        CLOSE H
        EXCHECK = 0
        KILL F$
    END IF
END FUNCTION

' Finds a file or files in the current directory...
FUNCTION FindFile (Name$(), Size$())

    ' Get current directory
    SHELL "CD>\SOKOBAN.DIR"
    OPEN "\SOKOBAN.DIR" FOR INPUT AS #1
    LINE INPUT #1, Name$(0)
    CLOSE #1
    IF LEN(Name$(0)) > 47 THEN Name$(0) = "..." + MID$(Name$(0), LEN(Name$(0)) - 45, 42)

    ' Read directory files
    SHELL "DIR/A/O>\SOKOBAN.DIR"
    OPEN "\SOKOBAN.DIR" FOR INPUT AS #1
    LINE INPUT #1, L$
    DO
        LINE INPUT #1, L$
        IF L$ = "" THEN EXIT DO
    LOOP

    DO WHILE NOT EOF(1)
        LINE INPUT #1, L$
        IF MID$(L$, 1, 1) = " " THEN EXIT DO
        Name$ = RTRIM$(MID$(L$, 1, 8))
        Extension$ = RTRIM$(MID$(L$, 10, 3))
        IF LEN(Extension$) > 0 THEN Name$ = Name$ + "." + Extension$
        IF NOT Name$ = "SOKOBAN.DIR" AND NOT Name$ = "SOKOBAN.TMP" AND NOT Name$ = "." THEN
            EC = 0
            I = I + 1
            Name$(I) = Name$
            Type$ = "  "
            IF NOT MID$(L$, 14, 5) = "<DIR>" THEN
                OPEN Name$ FOR INPUT AS #2
                IF INPUT$(10, 2) = "SOKOMAZE10" THEN Type$ = " *"
                CLOSE #2
            END IF
            Size$(I) = MID$(L$, 13, 30) + Type$
            IF EC > 0 THEN EXIT DO
        END IF
    LOOP
    CLOSE #1
    KILL "\SOKOBAN.DIR"
    FindFile = I
END FUNCTION

' Wait until a key is pressed, then return its value.
FUNCTION GETKEY$
    Key$ = ""
    DO WHILE Key$ = ""
        Key$ = INKEY$
    LOOP
    GETKEY$ = Key$
END FUNCTION

' Read a string from the keyboard...
FUNCTION GetText (Text$, MaxLength, Filter$, ESCAPE)

    IF MaxLength > 512 THEN MaxLength = 512
    IF LEN(Text$) > MaxLength THEN Text$ = MID$(Text$, 1, MaxLength)
    CursorCounter = LEN(Text$)
    Col = POS(0) + CursorCounter MOD 80
    Row = CSRLIN + CursorCounter / 80
    PRINT Text$;
    LOCATE , , 1, 12, 13
    Filter$ = UCASE$(Filter$)

    DO
        LOCATE Row, Col: PRINT ;
        Key$ = GETKEY$
        SELECT CASE Key$
            CASE CHR$(27):          ' [Esc]
                IF ESCAPE > 0 THEN
                    GetText = -1
                    LOCATE , , 0
                    EXIT FUNCTION
                END IF
            CASE CHR$(13):          ' [Enter]
                GetText = LEN(Text$)
                LOCATE , , 0
                EXIT FUNCTION
            CASE CHR$(0) + "G":     ' [Home]
                DO WHILE CursorCounter > 0
                    GOSUB CursorBackward
                LOOP
            CASE CHR$(0) + "O":     ' [End]
                TextLen = LEN(Text$)
                IF LEN(Text$) = MaxLength THEN TextLen = TextLen - 1
                DO WHILE CursorCounter < TextLen
                    GOSUB CursorForeward
                LOOP
            CASE CHR$(0) + "R":     ' [Insert]
                Insert = NOT Insert
                IF Insert = 0 THEN
                    LOCATE , , , 12, 13
                ELSE
                    LOCATE , , , 1, 13
                END IF
            CASE CHR$(0) + "S":     ' [Delete]
                IF LEN(Text$) > 0 THEN
                    LOCATE Row, Col
                    FOR Counter = CursorCounter + 2 TO LEN(Text$)
                        PRINT MID$(Text$, Counter, 1);
                    NEXT Counter
                    PRINT " ";
                    Text$ = MID$(Text$, 1, CursorCounter) + MID$(Text$, CursorCounter + 2, LEN(Text$))
                END IF
            CASE CHR$(8):           ' [BackSpace]
                IF LEN(Text$) > 0 THEN
                    GOSUB CursorBackward
                    LOCATE Row, Col
                    FOR Counter = CursorCounter + 2 TO LEN(Text$)
                        PRINT MID$(Text$, Counter, 1);
                    NEXT Counter
                    PRINT " ";
                    Text$ = MID$(Text$, 1, CursorCounter) + MID$(Text$, CursorCounter + 2, LEN(Text$))
                END IF
            CASE CHR$(0) + "M":     ' [Right Arrow]
                GOSUB CursorForeward
            CASE CHR$(0) + "K":     ' [Left Arrow]
                GOSUB CursorBackward
            CASE ELSE:
                IF NOT MID$(Key$, 1, 1) = CHR$(0) THEN
                    IF Filter$ = "" THEN
                        IF ASC(Key$) >= 32 AND ASC(Key$) <= 255 THEN GOSUB WriteCharacter
                    ELSE
                        FOR Counter = 1 TO LEN(Filter$)
                            IF UCASE$(Key$) = MID$(Filter$, Counter, 1) THEN GOSUB WriteCharacter: EXIT FOR
                        NEXT Counter
                    END IF
                END IF
        END SELECT
    LOOP
EXIT FUNCTION
'* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *'
WriteCharacter:
    IF Insert = 0 THEN
        LOCATE Row, Col: PRINT Key$;
        Text$ = MID$(Text$, 1, CursorCounter) + Key$ + MID$(Text$, CursorCounter + 2, LEN(Text$))
        GOSUB CursorForeward
    ELSEIF LEN(Text$) < MaxLength THEN
        LOCATE Row, Col: PRINT Key$;
        FOR Counter = CursorCounter + 1 TO LEN(Text$)
            PRINT MID$(Text$, Counter, 1);
        NEXT Counter
        Text$ = MID$(Text$, 1, CursorCounter) + Key$ + MID$(Text$, CursorCounter + 1, LEN(Text$))
        GOSUB CursorForeward
    END IF
RETURN
'* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *'
CursorForeward:
    IF MaxLength > CursorCounter + 1 AND CursorCounter < LEN(Text$) THEN
        IF Col = 80 THEN
            Col = 1
            IF Row < 24 THEN Row = Row + 1 ELSE PRINT
        ELSE
            Col = Col + 1
        END IF
        CursorCounter = CursorCounter + 1
    END IF
RETURN
'* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *'
CursorBackward:
    IF CursorCounter > 0 THEN
        IF Col = 1 THEN
            Col = 80
            Row = Row - 1
        ELSE
            Col = Col - 1
        END IF
        CursorCounter = CursorCounter - 1
    END IF
RETURN
'* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *'
END FUNCTION

SUB Initialize
    SCREEN 0
    CLS

    LOCATE 1, 1, 0, 12, 13
END SUB

' Load SOKOBAN text encoded maze file.
SUB MLOAD (File$)
    ClearMaze
    OPEN File$ FOR INPUT AS #1

    ' Load SOKOBAN maze file identification string
    LINE INPUT #1, L$

    ' Cause Syntax error if identification string could not be found.
    IF NOT MID$(L$, 1, 10) = "SOKOMAZE10" THEN
        COLOR 15, 4
        PutBox 10, 10, 70, 16, "ɻȼͺ "
        LOCATE 10: PRINTC " SMZ FILE IDENTIFICATION ERROR "
        LOCATE 12: PRINTC "SMZ ID string not found in file - " + File$
        COLOR 14
        LOCATE 14: PRINTC "Quit/Skip/Ignore (Q/S/I)?"
        Key$ = GETKEY$
        IF UCASE$(Key$) = "Q" OR Key$ = CHR$(27) THEN
            CX = -2
            CLOSE #1
            EXIT SUB
        ELSEIF NOT UCASE$(Key$) = "I" AND NOT Key$ = CHR$(13) THEN
            CX = -1
            CLOSE #1
            EXIT SUB
        END IF
    END IF

    ' Get character position.
    CX = ASC(MID$(L$, 11, 1)) - 33
    CY = ASC(MID$(L$, 12, 1)) - 33

    ' Load maze.
    FOR V = 1 TO 20
        IF EOF(1) > 0 THEN EXIT FOR
        LINE INPUT #1, L$
        L$ = LTRIM$(RTRIM$(L$))
        FOR H = 1 TO 40
            C = ASC(MID$(L$, H + 1, 1)) - 33
            MID$(MAZE$(V), H) = CHR$(C)
            SELECT CASE C
                CASE uWLL: MWC = MWC + 1
                CASE uBOX: MBC = MBC + 1
                CASE uTRG: MTC = MTC + 1
                CASE ELSE:
            END SELECT
        NEXT H
    NEXT V
    CLOSE #1
END SUB

' Save SOKOBAN maze as encoded text.
SUB MSAVE (File$)
    OPEN File$ FOR OUTPUT AS #1

    ' Save SOKOBAN maze file identification string and character position.
    PRINT #1, "SOKOMAZE10" + CHR$(CX + 33) + CHR$(CY + 33)

    ' Save maze.
    FOR V = 1 TO 20
        PRINT #1, "#";
        FOR H = 1 TO 40
            PRINT #1, CHR$(ASC(MID$(MAZE$(V), H, 1)) + 33);
        NEXT H
        PRINT #1, ""
    NEXT V
    CLOSE #1
END SUB

' Browse for maze files...
SUB OpenMaze (File$, CHANGE)
    DIM N$(1024)        ' File name
    DIM S$(1024)        ' File size, date, time, etc...
    DIM Buffer(4000)    ' Screen Buffer

    ' Save desktop
    ReadScreen Buffer()

    COLOR 15, 7
    PutWindow 15, 5, 65, 18, "ɻȼͺ*"
    LOCATE 5: PRINTC " OPEN MAZE "
    PutBox 17, 7, 63, 16, "ڿĳ"

    AF = 1
    DO
        IF F = 0 THEN
            COLOR 11, 3
            LOCATE 22, 1: PRINT " <Enter=Open> <Esc=Cancel> <Delete=Delete file> <Arrow=Next File>" + SPACE$(36);
            LOCATE 23, 1: PRINT SPACE$(80);
            COLOR 14, 7
            LOCATE 17, 17: PRINT "Reading current directory..."
            F = FindFile(N$(), S$())
            LOCATE 17, 17: PRINT SPACE$(30)
        END IF

        COLOR 15, 7
        LOCATE 6, 17: PRINT N$(0) + SPACE$(47 - LEN(N$(0)))

        FOR I = 1 TO 8
            LOCATE 7 + I, 18
            IF MID$(S$(I + S), 31, 1) = "*" THEN COLOR 10 ELSE COLOR 15
            IF AF = I + S THEN COLOR , 0 ELSE COLOR , 7
            IF I > F THEN PRINT SPACE$(45) ELSE PRINT " " + N$(I + S) + SPACE$(13 - LEN(N$(I + S))) + S$(I + S)
        NEXT I

        SELECT CASE GETKEY$
            CASE CHR$(27):          ' ESC
                EXIT DO
            CASE CHR$(13):          ' ENTER
                IF MID$(S$(AF), 2, 5) = "<DIR>" THEN
                    CHDIR N$(AF)
                    F = 0
                    S = 0
                    AF = 1
                ELSEIF MID$(S$(AF), 31, 1) = "*" THEN
                    File$ = N$(AF)
                    MLOAD File$

                    ' Chop file extension.
                    FOR I = 1 TO 8
                        IF MID$(File$, I, 1) = "." THEN EXIT FOR
                    NEXT I
                    File$ = MID$(File$, 1, I - 1)
                    CHANGE = 0
                    EXIT DO
                END IF
            CASE CHR$(0) + "S":     ' DELETE
                IF ASK("DELETE", "Are you sure, you want to delete " + N$(AF), 4) = 1 THEN
                    KILL N$(AF)
                    IF AF = F THEN AF = AF - 1
                    IF S > 0 THEN S = S - 1
                    F = 0
                END IF
            CASE CHR$(0) + "H":     ' UP
                IF AF > 1 THEN
                    AF = AF - 1
                    IF AF - S = 0 THEN S = S - 1
                END IF
            CASE CHR$(0) + "P":     ' DOWN
                IF AF < F THEN
                    AF = AF + 1
                    IF AF - S > 8 THEN S = S + 1
                END IF
            CASE CHR$(0) + "I":     ' PGUP
                S = S - 8
                IF S < 0 THEN S = 0
                AF = AF - 8
                IF AF < 1 THEN AF = 1
            CASE CHR$(0) + "Q":     ' PGDN
                S = S + 8
                IF S > F - 8 THEN S = F - 8
                AF = AF + 8
                IF AF > F THEN AF = F
            CASE CHR$(0) + "G":     ' HOME
                AF = 1
                S = 0
            CASE CHR$(0) + "O":     ' END
                AF = F
                IF F > 8 THEN S = F - 8
            CASE ELSE:
        END SELECT
    LOOP

    ' Restore desktop
    WriteScreen Buffer()

    DrawMaze
END SUB

' Print a string in the center.
SUB PRINTC (S$)
    LOCATE , 40.5 - LEN(S$) / 2
    PRINT S$;
END SUB

' Draw a box on the screen.
SUB PutBox (Col1, Row1, Col2, Row2, LineSet$)

    BoxWidth = Col2 - Col1 + 1
    LOCATE Row1, Col1: PRINT MID$(LineSet$, 1, 1) + STRING$(BoxWidth - 2, MID$(LineSet$, 5, 1)) + MID$(LineSet$, 2, 1);
    FOR I = Row1 + 1 TO Row2 - 1
        LOCATE I, Col1: PRINT MID$(LineSet$, 6, 1);
        IF LEN(LineSet$) > 6 THEN PRINT STRING$(BoxWidth - 2, MID$(LineSet$, 7, 1));
        LOCATE I, Col2: PRINT MID$(LineSet$, 6, 1);
    NEXT I
    LOCATE Row2, Col1: PRINT MID$(LineSet$, 3, 1) + STRING$(BoxWidth - 2, MID$(LineSet$, 5, 1)) + MID$(LineSet$, 4, 1);
END SUB

' Open a window on the screen.
SUB PutWindow (Col1, Row1, Col2, Row2, LineSet$)

    ' Draw Shadow
    IF LEN(LineSet$) > 6 AND NOT MID$(LineSet$, LEN(LineSet$), 1) = " " THEN
        DEF SEG = &HB800
        FOR I = Col1 TO Col2
            POKE Row2 * 160 + I * 2 + 3, 8
        NEXT I
        FOR I = Row1 TO Row2
            POKE I * 160 + Col2 * 2 + 1, 8
            POKE I * 160 + Col2 * 2 + 3, 8
        NEXT I
    END IF

    ' Draw Box
    PutBox Col1, Row1, Col2, Row2, MID$(LineSet$, 1, 6) + " "
END SUB

' Read the screen's contents from the memory into another block.
SUB ReadScreen (Buffer())

    DEF SEG = &HB800
    FOR I = 0 TO 4000
        Buffer(I) = PEEK(I)
    NEXT I
END SUB

' Save maze file...
SUB SaveMaze (File$, CHANGE)
    DIM Buffer(4000)    ' Screen Buffer

    ' Save desktop
    ReadScreen Buffer()

    DO UNTIL Key$ = "Y" OR Key$ = CHR$(13)
        COLOR 11, 3
        LOCATE 22, 1: PRINT " <Enter=Save> <Esc=Cancel>" + SPACE$(80)
        LOCATE 23, 1: PRINT SPACE$(80);
        COLOR , 0
        LOCATE 24, 1: PRINT SPACE$(80);
        LOCATE 25, 1: PRINT SPACE$(80);

        COLOR 15, 7
        PutWindow 10, 7, 70, 15, "ɻȼͺ*"
        COLOR 15, 7
        LOCATE 7: PRINTC " SAVE "
        LOCATE 11, 15: PRINT "Please enter the file name without the extension."
        LOCATE 12, 15: PRINT "It must be at least one character long,"
        LOCATE 13, 15: PRINT "but may not exceed 8 characters."
        LOCATE 9, 12: PRINT "File name: "; : COLOR 14, 1: PRINT SPACE$(45);
        LOCATE 9, 23
        IF GetText(File$, 8, "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789", 1) > 0 THEN
            File$ = UCASE$(File$)
            COLOR 11, 3: LOCATE 22, 1: PRINT SPACE$(80);
            IF EXCHECK(File$ + ".SMZ") > 0 THEN
                COLOR 14, 4
                LOCATE 14, 11: PRINT " " + File$ + ".SMZ already exists. Overwrite (Y/N)?" + SPACE$(21 - LEN(File$))
                Key$ = UCASE$(GETKEY$)
            ELSE
                Key$ = "Y"
            END IF
        ELSE
            EXIT DO
        END IF
    LOOP

    IF Key$ = "Y" THEN
        MSAVE File$ + ".SMZ"
        CHANGE = 0
    END IF

    ' Restore desktop
    WriteScreen Buffer()
END SUB

' Edit level...
SUB SokoEdit
    COLOR 15, 0
    CLS

    ClearMaze
    DrawMaze
    COLOR 15, 0: LOCATE 21, 1: PRINT STRING$(80, "");
    COLOR 11, 3
    LOCATE 22, 1: PRINT " <Insert=Insert Mode> <T=Put Box Target> <B=Put Box> <Esc=Exit>                 ";
    LOCATE 23, 1: PRINT " <Delete=Delete Mode> <Space=Put Spaces> <F6=Clear Maze> <F2=Save As> <F3=Open> ";

    CX = 20
    CY = 10
    File$ = "NONAME00"

    DO
        CU = ASC(MID$(MAZE$(CY), CX))   ' Current Unit

        COLOR , 0
        LOCATE 25, 74
        IF IM = 1 THEN          ' Insert Mode
            COLOR 10 + 16
            PRINT "INSERT";
            GOSUB PutWLL
        ELSEIF DM = 1 THEN      ' Delete Mode
            COLOR 12 + 16
            PRINT "DELETE";
            GOSUB PutSPC
        ELSE
            PRINT SPACE$(6);    ' INS & DEL OFF
        END IF

        CU = ASC(MID$(MAZE$(CY), CX))   ' Update Current Unit

        ' Draw new cell
        DrawCell CX, CY, ASC(MID$(MAZE$(CY), CX))

        ' Remember last position
        LCX = CX
        LCY = CY
        LCC = SCREEN(CY, CX * 2 - 1, 1)

        ' Paint character
        COLOR 15, 7
        LOCATE CY, CX * 2 - 1: PRINT CHR$(SCREEN(CY, CX * 2 - 1)) + CHR$(SCREEN(CY, CX * 2));

        ' Print counters
        COLOR 14, 1
        LOCATE 24, 1: PRINT "   " + CHR$(26) + "WLL:"; MWC,
        LOCATE 25, 1: PRINT "   " + CHR$(26) + "BOX:"; MBC,
        LOCATE 24, 15: PRINT "   " + CHR$(26) + "TRG:"; MTC, "   ";
        LOCATE 25, 15: PRINT "   " + CHR$(26) + "SPC:"; 800 - MWC - MBC - MTC, "   ";

        ' Print file name & CHANGE signal
        COLOR 15, 5
        LOCATE 25, 33: PRINT " File: " + MID$(File$, 1, 8) + ".SMZ" + SPACE$(13 - LEN(MID$(File$, 1, 8)));
        COLOR 11
        IF CHANGE = 0 THEN PRINT "  ";  ELSE PRINT ; "* ";


        Key$ = GETKEY$

        IF UCASE$(Key$) = "T" OR UCASE$(Key$) = "X" THEN    ' Put one TARGET
            GOSUB PutTRG
        ELSEIF UCASE$(Key$) = "B" THEN                      ' Put one BOX
            GOSUB PutBox
        ELSEIF UCASE$(Key$) = "W" THEN                      ' Put one WALL
            DM = 0  ' Delete Mode OFF
            IM = 0  ' Insert Mode OFF
            GOSUB PutWLL
        ELSEIF Key$ = " " THEN                              ' DELETE one unit
            DM = 0  ' Delete Mode OFF
            IM = 0  ' Insert Mode OFF
            GOSUB PutSPC
        END IF

        SELECT CASE Key$
            CASE CHR$(0) + "<":     ' F2
                DM = 0  ' Delete Mode OFF
                IM = 0  ' Insert Mode OFF
                SaveMaze File$, CHANGE
            CASE CHR$(0) + "=":     ' F3
                DM = 0  ' Delete Mode OFF
                IM = 0  ' Insert Mode OFF
                IF CHANGE = 1 THEN GOSUB AskSave
                OpenMaze File$, CHANGE
                LCX = 0
                LCY = 0
            CASE CHR$(0) + "@":     ' F6
                IF ASK("CLEAR MAZE", "Are you sure, you want to clear the maze", 7) = 1 THEN
                    LCX = 0
                    LCY = 0
                    CX = 20
                    CY = 10
                    DM = 0  ' Delete Mode OFF
                    IM = 0  ' Insert Mode OFF
                    FOR Y = 1 TO 20
                        FOR x = 1 TO 40
                            IF ASC(MID$(MAZE$(Y), x, 1)) > 0 THEN CHANGE = 1
                        NEXT x
                    NEXT Y
                    ClearMaze
                END IF
                DrawMaze
            CASE CHR$(0) + "R":     ' INSERT
                DM = 0
                IF IM = 0 THEN IM = 1 ELSE IM = 0
            CASE CHR$(0) + "S":     ' DELETE
                IM = 0
                IF DM = 0 THEN DM = 1 ELSE DM = 0
            CASE CHR$(0) + "H":     ' UP
                IF CY > 1 THEN CY = CY - 1 ELSE CY = 20
            CASE CHR$(0) + "P":     ' DOWN
                IF CY < 20 THEN CY = CY + 1 ELSE CY = 1
            CASE CHR$(0) + "K":     ' LEFT
                IF CX > 1 THEN CX = CX - 1 ELSE CX = 40
            CASE CHR$(0) + "M":     ' RIGHT
                IF CX < 40 THEN CX = CX + 1 ELSE CX = 1
            CASE CHR$(27):          ' ESC
                IF CHANGE = 1 THEN
                    COLOR 11, 3
                    LOCATE 22, 1: PRINT SPACE$(80);
                    LOCATE 23, 1: PRINT SPACE$(80);
                    COLOR 14 + 16, 0
                    LOCATE 24, 1: PRINT SPACE$(80);
                    COLOR , 4
                    LOCATE 25, 1: PRINT SPACE$(80);
                    LOCATE 25: PRINTC "<<< YOU ARE ABOUT TO EXIT >>>"
                    GOSUB AskSave
                END IF
                EXIT DO
            CASE ELSE:
        END SELECT

        ' Redraw previous unit
        DEF SEG = &HB800
        POKE (LCY - 1) * 160 + (LCX - 1) * 4 + 1, LCC
        POKE (LCY - 1) * 160 + (LCX - 1) * 4 + 3, LCC

    LOOP
EXIT SUB
'* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *'
AskSave:
    IF ASK("SAVE", "Would you like to save changes", 7) = 1 THEN SaveMaze File$, CHANGE
    DrawMaze
RETURN
'* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *'
PutSPC:

    IF NOT CU = uSPC THEN
        CHANGE = 1
        MID$(MAZE$(CY), CX) = CHR$(uSPC)

        ' Update counters
        IF CU = uWLL THEN MWC = MWC - 1
        IF CU = uBOX THEN MBC = MBC - 1
        IF CU = uTRG THEN MTC = MTC - 1
    END IF
RETURN
'* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *'
PutWLL:
    IF NOT CU = uWLL THEN
        CHANGE = 1
        MID$(MAZE$(CY), CX) = CHR$(uWLL)

        ' Update counters
        MWC = MWC + 1
        IF CU = uTRG THEN MTC = MTC - 1
        IF CU = uBOX THEN MBC = MBC - 1
    END IF
RETURN
'* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *'
PutBox:
    DM = 0  ' Delete Mode OFF
    IM = 0  ' Insert Mode OFF

    IF NOT CU = uBOX THEN
        CHANGE = 1
        MID$(MAZE$(CY), CX) = CHR$(uBOX)

        ' Update counters
        MBC = MBC + 1
        IF CU = uWLL THEN MWC = MWC - 1
        IF CU = uTRG THEN MTC = MTC - 1
    END IF
RETURN
'* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *'
PutTRG:
    DM = 0  ' Delete Mode OFF
    IM = 0  ' Insert Mode OFF

    IF NOT CU = uTRG THEN
        CHANGE = 1
        MID$(MAZE$(CY), CX) = CHR$(uTRG)

        ' Update counters
        MTC = MTC + 1
        IF CU = uWLL THEN MWC = MWC - 1
        IF CU = uBOX THEN MBC = MBC - 1
    END IF
RETURN
'* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *'
END SUB

' SOKOBAN introduction & Main Menu function
FUNCTION SokoMenu
    SCREEN 0
    LOCATE , , 0, 12, 13
    WIDTH 80, 25
    COLOR 15, 1
    CLS

    PutBox 1, 1, 80, 25, "ɻȼͺ"

    COLOR 11
    LOCATE 3, 67: PRINT "Version 1A"
    LOCATE 3, 5: PRINT "Welcome to SOKOBAN - The Maze Game!"
    LOCATE 4, 5: PRINT "Written by Zsolt Nagy Perge in Dec. 1999.";
    COLOR 15
    LOCATE 7, 5: PRINT "The object of this game is to push the boxes on their places."
    LOCATE 8, 5: PRINT "The boxes cannot be pulled, only pushed"
    LOCATE 9, 5: PRINT "and only one box can be pushed by a move."
    LOCATE 10, 5: PRINT "Once you pushed all the boxes, you can go on to the next level."

    AM = 1
    PutBox 25, 14, 55, 18, "ڿĳ"
    PutBox 23, 13, 57, 19, "ڿĳ"
    LOCATE 13: PRINTC " Main Menu "
    DO
        FOR I = 1 TO 3
            IF AM = I THEN COLOR 14, 5 ELSE COLOR 14, 1
            LOCATE 14 + I, 26
            SELECT CASE I
                CASE 1: PRINT " Start Game...               "
                CASE 2: PRINT " Edit Level...               "
                CASE 3: PRINT " Exit                        "
                CASE ELSE:
            END SELECT
        NEXT I

        SELECT CASE GETKEY$
            CASE CHR$(13):          ' ENTER
                EXIT DO
            CASE CHR$(0) + "H":     ' UP
                IF AM = 1 THEN AM = 3 ELSE AM = AM - 1
            CASE CHR$(0) + "P":     ' DOWN
                IF AM = 3 THEN AM = 1 ELSE AM = AM + 1
            CASE CHR$(27):          ' ESC
                AM = 3
                EXIT DO
            CASE ELSE:
        END SELECT
    LOOP

    SokoMenu = AM
END FUNCTION

' Select levels and play...
SUB SokoPlay
    L = 1
    SHELL "DIR.SMZ/O/B>SOKOBAN.TMP"
    DO

        ' Select level...
        COLOR 15, 0
        CLS

        I = 0
        OPEN "SOKOBAN.TMP" FOR INPUT AS #1
        DO WHILE NOT EOF(1)
            I = I + 1
            LINE INPUT #1, L$
            IF L = I THEN File$ = L$
        LOOP
        CLOSE #1

        ' Exit, if no SMZ maze files were found...
        IF I = 0 THEN
            COLOR 15, 4
            PutBox 5, 10, 75, 16, "ɻȼͺ "
            LOCATE 10: PRINTC " ERROR "
            LOCATE 14, 6: PRINT STRING$(69, "")
            LOCATE 15: PRINTC "< PRESS ANY KEY TO RETURN >"
            COLOR 14
            LOCATE 12: PRINTC "SOKOBAN could not find any SMZ maze files in the current directory."
            K$ = GETKEY$
            EXIT DO
        END IF

        ' Load maze file...
        MLOAD File$
        IF CX = -2 THEN EXIT SUB    ' USER CHOSE TO QUIT FROM SOKOPLAY
        IF CX = -1 THEN L = L + 1   ' USER CHOSE TO SKIP THIS LEVEL
        IF CX > 0 THEN              ' CHECK ID STRING...
            DrawMaze

            ' Print file name
            COLOR 15, 5
            LOCATE 25, 33: PRINT " File: " + MID$(File$, 1, 12) + SPACE$(18 - LEN(MID$(File$, 1, 8)));

            ' Print counters
            COLOR 10, 1
            LOCATE 24, 1: PRINT "   " + CHR$(26) + "WLL:"; MWC,
            LOCATE 25, 1: PRINT "   " + CHR$(26) + "BOX:"; MBC,
            LOCATE 24, 15: PRINT "   " + CHR$(26) + "TRG:"; MTC, "   ";
            LOCATE 25, 15: PRINT "   " + CHR$(26) + "SPC:"; 800 - MWC - MBC - MTC, "   ";

            ' Skip level, if maze is incorrect; otherwise, play the game...
            IF MID$(MAZE$(CY), CX, 1) = CHR$(1) OR NOT MBC = MTC THEN
                COLOR 15, 4
                PutBox 20, 7, 60, 13, "ɻȼͺ "
                LOCATE 7: PRINTC " MAZE ERROR "
                LOCATE 11, 21: PRINT STRING$(39, "")
                LOCATE 12: PRINTC "< PRESS ANY KEY TO SKIP THIS LEVEL >"
                COLOR 14
                LOCATE 9: PRINTC "Incorrect maze file."
                K$ = GETKEY$
                L = L + 1
            ELSE
                COLOR 15, 0: LOCATE 21, 1: PRINT STRING$(80, "");
                COLOR 11, 3
                LOCATE 22, 1: PRINT " <F10=Restart Level> <Esc=Exit>" + SPACE$(49);
                LOCATE 23, 1: PRINT SPACE$(80);

                DO
                    ' Remember last position
                    LCX = CX
                    LCY = CY
                    LCC = SCREEN(CY, CX * 2 - 1, 1)

                    ' Paint character
                    COLOR 15, 7
                    LOCATE CY, CX * 2 - 1: PRINT CHR$(SCREEN(CY, CX * 2 - 1)) + CHR$(SCREEN(CY, CX * 2));

                    LOCATE CY, CX * 2 - 1
                    COLOR 15, 7
                    PRINT CHR$(SCREEN(CY, CX * 2 - 1)) + CHR$(SCREEN(CY, CX * 2))

                    ' Check if done...
                    DONE = 0
                    FOR V = 1 TO 20
                        FOR H = 1 TO 40
                            IF MID$(MAZE$(V), H, 1) = CHR$(uBOX) THEN DONE = 1
                        NEXT H
                    NEXT V
                    IF DONE = 0 THEN EXIT DO

                    ' Print current position
                    COLOR 14, 3
                    LOCATE 25, 63: PRINT "  X ="; CX;
                    LOCATE 25, 71: PRINT "  Y ="; CY; " ";

                    Key$ = GETKEY$
                    SELECT CASE Key$
                        CASE CHR$(0) + "D":     ' F10
                            IF ASK("RESTART", "Would you like to start this level again", 2) = 1 THEN
                                ' Reload maze file...
                                MLOAD File$
                                DrawMaze
                                LCX = 0
                                LCY = 0
                            END IF
                        CASE CHR$(0) + "H":     ' UP
                            IF CY > 1 THEN
                                M1 = ASC(MID$(MAZE$(CY - 1), CX, 1))
                                IF CY > 2 THEN M2 = ASC(MID$(MAZE$(CY - 2), CX, 1))
                                IF M1 = uSPC OR M1 = uTRG THEN
                                    CY = CY - 1
                                ELSEIF M1 = uBOX AND M2 = uSPC THEN
                                    MID$(MAZE$(CY - 1), CX, 1) = CHR$(uSPC)
                                    MID$(MAZE$(CY - 2), CX, 1) = CHR$(uBOX)
                                    CY = CY - 1
                                ELSEIF M1 = uBOT AND M2 = uSPC THEN
                                    MID$(MAZE$(CY - 1), CX, 1) = CHR$(uTRG)
                                    MID$(MAZE$(CY - 2), CX, 1) = CHR$(uBOX)
                                    CY = CY - 1
                                ELSEIF M1 = uBOX AND M2 = uTRG THEN
                                    MID$(MAZE$(CY - 1), CX, 1) = CHR$(uSPC)
                                    MID$(MAZE$(CY - 2), CX, 1) = CHR$(uBOT)
                                    CY = CY - 1
                                ELSEIF M1 = uBOT AND M2 = uTRG THEN
                                    MID$(MAZE$(CY - 1), CX, 1) = CHR$(uTRG)
                                    MID$(MAZE$(CY - 2), CX, 1) = CHR$(uBOT)
                                    CY = CY - 1
                                END IF
                                IF CY > 2 THEN DrawCell CX, CY - 2, ASC(MID$(MAZE$(CY - 2), CX, 1))
                                DrawCell CX, CY - 1, ASC(MID$(MAZE$(CY - 1), CX, 1))
                            END IF
                        CASE CHR$(0) + "P":     ' DOWN
                            IF CY < 20 THEN
                                M1 = ASC(MID$(MAZE$(CY + 1), CX, 1))
                                IF CY < 19 THEN M2 = ASC(MID$(MAZE$(CY + 2), CX, 1))
                                IF M1 = uSPC OR M1 = uTRG THEN
                                    CY = CY + 1
                                ELSEIF M1 = uBOX AND M2 = uSPC THEN
                                    MID$(MAZE$(CY + 1), CX, 1) = CHR$(uSPC)
                                    MID$(MAZE$(CY + 2), CX, 1) = CHR$(uBOX)
                                    CY = CY + 1
                                ELSEIF M1 = uBOT AND M2 = uSPC THEN
                                    MID$(MAZE$(CY + 1), CX, 1) = CHR$(uTRG)
                                    MID$(MAZE$(CY + 2), CX, 1) = CHR$(uBOX)
                                    CY = CY + 1
                                ELSEIF M1 = uBOX AND M2 = uTRG THEN
                                    MID$(MAZE$(CY + 1), CX, 1) = CHR$(uSPC)
                                    MID$(MAZE$(CY + 2), CX, 1) = CHR$(uBOT)
                                    CY = CY + 1
                                ELSEIF M1 = uBOT AND M2 = uTRG THEN
                                    MID$(MAZE$(CY + 1), CX, 1) = CHR$(uTRG)
                                    MID$(MAZE$(CY + 2), CX, 1) = CHR$(uBOT)
                                    CY = CY + 1
                                END IF
                                IF CY < 19 THEN DrawCell CX, CY + 2, ASC(MID$(MAZE$(CY + 2), CX, 1))
                                DrawCell CX, CY + 1, ASC(MID$(MAZE$(CY + 1), CX, 1))
                            END IF
                        CASE CHR$(0) + "K":     ' LEFT
                            IF CX > 1 THEN
                                M1 = ASC(MID$(MAZE$(CY), CX - 1, 1))
                                IF CX > 2 THEN M2 = ASC(MID$(MAZE$(CY), CX - 2, 1))
                                IF M1 = uSPC OR M1 = uTRG THEN
                                    CX = CX - 1
                                ELSEIF M1 = uBOX AND M2 = uSPC THEN
                                    MID$(MAZE$(CY), CX - 1, 1) = CHR$(uSPC)
                                    MID$(MAZE$(CY), CX - 2, 1) = CHR$(uBOX)
                                    CX = CX - 1
                                ELSEIF M1 = uBOT AND M2 = uSPC THEN
                                    MID$(MAZE$(CY), CX - 1, 1) = CHR$(uTRG)
                                    MID$(MAZE$(CY), CX - 2, 1) = CHR$(uBOX)
                                    CX = CX - 1
                                ELSEIF M1 = uBOX AND M2 = uTRG THEN
                                    MID$(MAZE$(CY), CX - 1, 1) = CHR$(uSPC)
                                    MID$(MAZE$(CY), CX - 2, 1) = CHR$(uBOT)
                                    CX = CX - 1
                                ELSEIF M1 = uBOT AND M2 = uTRG THEN
                                    MID$(MAZE$(CY), CX - 1, 1) = CHR$(uTRG)
                                    MID$(MAZE$(CY), CX - 2, 1) = CHR$(uBOT)
                                    CX = CX - 1
                                END IF
                                IF CX > 2 THEN DrawCell CX - 2, CY, ASC(MID$(MAZE$(CY), CX - 2, 1))
                                DrawCell CX - 1, CY, ASC(MID$(MAZE$(CY), CX - 1, 1))
                            END IF
                        CASE CHR$(0) + "M":     ' RIGHT
                            IF CX < 40 THEN
                                M1 = ASC(MID$(MAZE$(CY), CX + 1, 1))
                                IF CX < 39 THEN M2 = ASC(MID$(MAZE$(CY), CX + 2, 1))
                                IF M1 = uSPC OR M1 = uTRG THEN
                                    CX = CX + 1
                                ELSEIF M1 = uBOX AND M2 = uSPC THEN
                                    MID$(MAZE$(CY), CX + 1, 1) = CHR$(uSPC)
                                    MID$(MAZE$(CY), CX + 2, 1) = CHR$(uBOX)
                                    CX = CX + 1
                                ELSEIF M1 = uBOT AND M2 = uSPC THEN
                                    MID$(MAZE$(CY), CX + 1, 1) = CHR$(uTRG)
                                    MID$(MAZE$(CY), CX + 2, 1) = CHR$(uBOX)
                                    CX = CX + 1
                                ELSEIF M1 = uBOX AND M2 = uTRG THEN
                                    MID$(MAZE$(CY), CX + 1, 1) = CHR$(uSPC)
                                    MID$(MAZE$(CY), CX + 2, 1) = CHR$(uBOT)
                                    CX = CX + 1
                                ELSEIF M1 = uBOT AND M2 = uTRG THEN
                                    MID$(MAZE$(CY), CX + 1, 1) = CHR$(uTRG)
                                    MID$(MAZE$(CY), CX + 2, 1) = CHR$(uBOT)
                                    CX = CX + 1
                                END IF
                                IF CX < 39 THEN DrawCell CX + 2, CY, ASC(MID$(MAZE$(CY), CX + 2, 1))
                                DrawCell CX + 1, CY, ASC(MID$(MAZE$(CY), CX + 1, 1))
                            END IF
                        CASE CHR$(27):          ' ESC
                            IF ASK("EXIT", "Do you want to exit", 7) = 1 THEN EXIT SUB
                        CASE ELSE:
                    END SELECT

                    DrawCell CX, CY, ASC(MID$(MAZE$(CY), CX, 1))

                    ' Redraw previous unit
                    DEF SEG = &HB800
                    POKE (LCY - 1) * 160 + (LCX - 1) * 4 + 1, LCC
                    POKE (LCY - 1) * 160 + (LCX - 1) * 4 + 3, LCC

                LOOP


                ' This level was completed...
                SLEEP 1
                LC = LC + 1
                COLOR 15, 5
                PutBox 6, 10, 74, 16, "ɻȼͺ "
                LOCATE 10: PRINTC " CONGRATULATIONS "
                LOCATE 14, 7: PRINT STRING$(67, "")
                LOCATE 15: PRINTC "< PRESS ENTER TO CONTINUE >"
                COLOR 14
                LOCATE 12: PRINTC "You have completed level" + STR$(LC) + " !"
                DO UNTIL GETKEY$ = CHR$(13)
                LOOP
                L = L + 1
            END IF
        END IF

        ' Exit, if all levels were completed...
        IF L > I THEN EXIT DO
    LOOP
END SUB

' Write a block from the memory on the screen.
SUB WriteScreen (Buffer())

    DEF SEG = &HB800
    FOR I = 0 TO 4000
        POKE I, Buffer(I)
    NEXT I
END SUB

